

#include <stdint.h>


using namespace std;


extern "C" {
	extern void (*__CTOR_LIST__)();
	extern void (*__DTOR_LIST__)();
	extern void (*__init_array_start)();
	extern void (*__init_array_end)();
	extern void (*__fini_array_start)();
	extern void (*__fini_array_end)();
}


void _callInitArray() {
	void (**f)() = &__init_array_start;
	while (f != &__init_array_end) {
		(*f)();
		f++;
	}
}


void _callConstructors() {
	void (**constructor)() = &__CTOR_LIST__;
	uint32_t total = *(uint32_t *) constructor;
	constructor++;
	while (total) {
		(*constructor)();
		total--;
		constructor++;
	}
}


void _callDestructors() {
	void (**destructor)() = &__DTOR_LIST__;
	uint32_t total = *(uint32_t *) destructor;
	destructor++;
	while (total) {
		(*destructor)();
		total--;
		destructor++;
	}
}


void _callFiniArray() {
	void (**f)() = &__fini_array_start;
	while (f != &__fini_array_end) {
		(*f)();
		f++;
	}
}


#define  RCU_CTL    *((uint32_t *) 0x40021000)
#define  RCU_CFG0   *((uint32_t *) 0x40021004)
#define  RCU_CFG1   *((uint32_t *) 0x4002102C)
#define  FLASH_ACR  *((uint32_t *) 0x40022000)


void _initClock() {
	// CK_SYS = 96 MHz
	// (up to 108 MHz but 96 MHz is more affordable because of 96 / 2 = 48 MHz for the USB)
	// External crystal is 8 MHz: (8 MHz / 1) * 12 = 96 MHz
	// PREDV0SEL = 0, PREDV0 = 1, PLLSEL = 1, PLLMF = 12, SCS = 0b10
	// USB prescaler: /2: 96 / 2 = 48 MHz
	// AHB prescaler: /1: 96 MHz   (so CPU mtime/mtimecmp clock at 96 / 4 = 24 MHz)
	// APB2 prescaler: /1: 96 MHz
	// APB1 prescaler: /2: 48 MHz
	// ADC prescaler: /8: 12 MHz
	//         USBFSPSC                 PLLMF                     PLLSEL                   APB2 prescaler
	RCU_CFG0 = (((uint32_t) 3) << 22) | (((uint32_t) 10) << 18) | (((uint32_t) 1) << 16) | (((uint32_t) 4) << 8);
	// HXTALEN = 1
	RCU_CTL = ((uint32_t) 1) << 16;
	// wait for xtal to be ready
	while ((RCU_CTL & (((uint32_t) 1) << 17)) == 0)
		;
	// enable pll
	RCU_CTL |= ((uint32_t) 1) << 24;
	// wait for pll to be locked
	while ((RCU_CTL & (((uint32_t) 1) << 25)) == 0)
		;
	// when pll locked, the select that clock as system clock
	RCU_CFG0 &= ~((uint32_t) 3);
	RCU_CFG0 |= 2;
	// wait until system clock selected
	while (((RCU_CFG0 >> 2) & 3) != 2)
		;
}


extern int main();


extern "C" {
	extern unsigned char flash_sdata;
	extern unsigned char ram_sdata;
	extern unsigned char ram_edata;
}


void _initDataRAM() {
	// init .data section (global variables) with flash data
	unsigned char *from = &flash_sdata;
	unsigned char *to = &ram_sdata;
	while (to != &ram_edata) {
		*to = *from;
		from++;
		to++;
	}
}


extern "C" {
	extern unsigned char ram_sbssdata;
	extern unsigned char ram_ebssdata;
}


void _initBssRAM() {
	// init .bss section with zeros
	unsigned char *p = &ram_sbssdata;
	while (p != &ram_ebssdata) {
		*p = 0;
		p++;
	}
}


void _startup_0() __attribute__((section(".startup0"), naked));   // startup located at begining of flash
void _startup_1() __attribute__((section(".startup1"), naked));   // rest of startup


void _startup_0() {
	asm volatile (
		"j %0" : : "i" (_startup_1)
	);
}


void _startup_1() {
	asm volatile (
		"la sp, 0x20005000"     // point SP to the end of SRAM
	);
	_initClock();
	_initDataRAM();
	_initBssRAM();
	_callConstructors();
	_callInitArray();
	main();
	_callFiniArray();
	_callDestructors();
	while (true)
		;
}


extern "C" void __cxa_pure_virtual() {}
void *__dso_handle = 0;
extern "C" void __cxa_atexit() {}

